# Generated by Django 3.0.6 on 2020-05-21 14:00

from django.db import migrations, models

import django_stripe.db.fields

from ..compat import JSONField


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="WebhookEvent",
            fields=[
                (
                    "db_created_at",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Database created at"
                    ),
                ),
                (
                    "db_last_updated_at",
                    models.DateTimeField(
                        auto_now=True, verbose_name="Database last updated at"
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "stripe_id",
                    django_stripe.db.fields.StripeIdField(max_length=255, unique=True),
                ),
                (
                    "mode",
                    models.CharField(
                        choices=[("LIVE", "Live"), ("TEST", "Test")], max_length=20
                    ),
                ),
                (
                    "state",
                    models.CharField(
                        choices=[("RECEIVED", "Received"), ("PROCESSED", "Processed")],
                        default="RECEIVED",
                        max_length=20,
                    ),
                ),
                ("event_type", models.CharField(max_length=100)),
                (
                    "data",
                    JSONField(
                        help_text=(
                            "The contents of data.object, representing the object the "
                            "webhook event was triggered for."
                        )
                    ),
                ),
                ("api_version", models.CharField(max_length=40)),
                ("stripe_created_at", models.DateTimeField()),
                ("request_id", models.CharField(blank=True, max_length=255)),
                (
                    "request_idempotency_key",
                    models.CharField(blank=True, max_length=255),
                ),
                ("headers", JSONField(blank=True)),
                (
                    "remote_ip",
                    models.GenericIPAddressField(
                        help_text="IP address of the remote client.",
                        verbose_name="Remote IP",
                    ),
                ),
            ],
            options={"abstract": False},
        )
    ]



"""
Webhook call handling.

Any function can register itself to handle a specific event
by using the @stripe_webhook_handler("event.name") syntax.

We deliberately maintain a 1:1 mapping between full event type
name and a function handler to keep things simple; it means that
given any full event name you can find all the handlers with a
grep.

It also means that if Stripe introduce a new sub-event, for a
parent event - then our handlers will not start automatically
receiving events for that new event type. This is a deliberate
design choice and differs from how djstripe handles handler
registration - so take care if migrating.

Do not rely on your handlers being called in the same order every
single time, as registration occurs on load - and it will be hard
for you to maintain any semblence of control. If you need actions
to happen in a specific order, then register one parent handler
and manage the code flow via that.

Usage:

    from django_stripe.models import WebhookEvent
    from django_stripe.webhooks import stripe_webhook_handler

    # Single event handler
    @stripe_webhook_handler("customer.subscription.deleted")
    def delete_customer_subscription(event: WebhookEvent) -> Any:
        pass

    # Multiple event handler
    @stripe_webhook_handler(
        "customer.subscription.created",
        "customer.subscription.deleted",
        "customer.subscription.updated",
    )
    def customer_handler(event: WebhookEvent) -> Any:
        pass
"""
from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Callable, List

if TYPE_CHECKING:
    from .models import WebhookEvent


__all__ = ["stripe_webhook_handler", "run_handlers"]


REGISTRY: defaultdict = defaultdict(list)


def stripe_webhook_handler(*event_types: List[str]) -> Callable:
    """
    Decorate a function to register it as a Stripe webhook handler.

    Functions must be registered against the full Stripe event name
    (e.g. 'customer.subscription.deleted') that they wish to handle.
    To handle multiple, pass extra full event type names as arguments.
    """

    def decorator(func: Callable) -> Callable:
        for event_type in event_types:
            REGISTRY[event_type].append(func)
        return func

    return decorator


def run_handlers(event: WebhookEvent) -> None:
    """
    Invoke all handlers registered against a full event type name.

    Handlers for each event type group are invoked in order of
    registration and as that cannot be guaranteed easily - the
    handlers should not be order-dependent, see the module note.
    """
    handlers = REGISTRY.get(event.event_type, [])
    if not handlers:
        return

    for handler_func in handlers:
        handler_func(event=event)

    event.mark_processed()
    event.save(update_fields=["state", "db_last_updated_at"])



import stripe
from django.apps import AppConfig as DjangoAppConfig

__version__ = "0.2.2"

default_app_config = "django_stripe.AppConfig"


class AppConfig(DjangoAppConfig):

    name = "django_stripe"
    verbose_name = "Stripe"

    def ready(self) -> None:
        from . import settings

        # Stripe library setup
        stripe.api_key = settings.SECRET_KEY
        stripe.api_version = settings.API_VERSION

        # Identify the plugin to Stripe
        stripe.set_app_info(
            "django-stripe-lite",
            version=__version__,
            url="https://github.com/yunojuno/django-stripe-lite",
        )



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from .models import WebhookEvent


class WebhookEventAdmin(admin.ModelAdmin):
    date_hierarchy = "stripe_created_at"
    list_display = (
        "stripe_id",
        "event_type",
        "state",
        "mode",
        "api_version",
        "remote_ip",
        "stripe_created_at",
        "db_created_at",
    )
    list_filter = ("state", "mode", "api_version", "event_type")
    fieldsets = (
        (None, {"fields": ("stripe_id", "event_type", "mode", "state")}),
        (
            _("Webhook Payload"),
            {"fields": ("data", "api_version", "stripe_created_at")},
        ),
        (
            _("Instigating Request Details"),
            {"fields": ("request_id", "request_idempotency_key")},
        ),
        (_("Webhook Request Details"), {"fields": ("headers", "remote_ip")}),
        (_("Database Timestamps"), {"fields": ("db_created_at", "db_last_updated_at")}),
    )
    readonly_fields = ("db_created_at", "db_last_updated_at")
    search_fields = ("event_type",)
    ordering = ("-db_created_at",)


admin.site.register(WebhookEvent, WebhookEventAdmin)



from __future__ import annotations

import logging

from django.db import models
from django.utils.translation import gettext_lazy as _

from ..db.fields import StripeIdField

logger = logging.getLogger(__name__)


class TimestampedModel(models.Model):
    db_created_at = models.DateTimeField(_("Database created at"), auto_now_add=True)
    db_last_updated_at = models.DateTimeField(
        _("Database last updated at"), auto_now=True
    )

    class Meta:
        abstract = True


class StripeModel(TimestampedModel):
    """Provide base functionality for all Stripe-object models."""

    class Mode(models.TextChoices):
        LIVE = ("LIVE", _("Live"))
        TEST = ("TEST", _("Test"))

        @classmethod
        def from_stripe_livemode(cls, livemode: bool) -> StripeModel.Mode:
            return cls.LIVE if livemode else cls.TEST  # type: ignore

    id = models.BigAutoField(primary_key=True)
    stripe_id = StripeIdField()

    mode = models.CharField(max_length=20, choices=Mode.choices)

    class Meta:
        abstract = True

    @property
    def is_live_mode(self) -> bool:
        return self.mode == self.Mode.LIVE

    @property
    def is_test_mode(self) -> bool:
        return self.mode == self.Mode.TEST



from __future__ import annotations

from typing import Any

from django.db import models


class StripeIdField(models.CharField):
    """Represent an identifier in Stripe format."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Provide sane defaults for fields storing Stripe IDs.

        See: https://stripe.com/docs/upgrades

        "You can safely assume object IDs we generate will
        never exceed 255 characters, but you should be able
        to handle IDs of up to that length."
        """
        defaults = {"max_length": 255, "blank": False, "null": False, "unique": True}
        defaults.update(kwargs)
        super().__init__(*args, **defaults)



import stripe
from django.conf import settings

# Determines whether Stripe API requests are sent as
# test or live requests to Stripe's systems.
LIVE_MODE = getattr(settings, "STRIPE_LIVE_MODE", False)

# Exportable keys
PUBLISHABLE_KEY = getattr(settings, "STRIPE_PUBLISHABLE_KEY", "")
SECRET_KEY = getattr(settings, "STRIPE_SECRET_KEY", "")

# Webhook secret: used for signature validation.
WEBHOOK_SECRET = getattr(settings, "STRIPE_WEBHOOK_SECRET", "")

# Webhook tolerance: the time window in seconds for which the
# inbound webhook event's signature should be consided valid
# for. Set to None to accept any timestamps (unadvisable).
WEBHOOK_TOLERANCE_SECONDS = getattr(
    settings, "STRIPE_WEBHOOK_TOLERANCE_SECONDS", stripe.Webhook.DEFAULT_TOLERANCE
)

API_VERSION = getattr(settings, "STRIPE_API_VERSION", "2020-03-02")



from __future__ import annotations

import logging

import stripe
from django.db import transaction
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
)
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from . import settings, webhooks
from .models import WebhookEvent

logger = logging.getLogger(__name__)


@require_POST
@csrf_exempt
@transaction.atomic
def webhook_view(request: HttpRequest) -> HttpResponse:
    """
    Handle inbound Stripe webhook events.

    This endpoint is open to the web, but we only accept requests
    that have a valid signature from Stripe themselves, based on
    our WEBHOOK_SECRET.

    Successly received => HTTP 200
    Invalid payload (as deemed by Stripe library) => HTTP 400
    Invalid/missing signature => HTTP 403
    Unhandled exception => HTTP 500

    Anything but an HTTP 200 response to Stripe will result in
    Stripe replaying the webhook at a later date. If the failures
    continue, Stripe will eventually put the webhook in the errored
    state and stop sending events.
    """
    try:
        event = stripe.Webhook.construct_event(
            request.body,
            request.headers.get("Stripe-Signature"),
            settings.WEBHOOK_SECRET,
        )
    except stripe.error.SignatureVerificationError:
        return HttpResponseForbidden("Invalid signature")
    except ValueError:
        return HttpResponseBadRequest("Invalid payload")

    persisted_event = WebhookEvent.objects.create_from_event(request, event)
    webhooks.run_handlers(persisted_event)
    return HttpResponse(persisted_event.id, status=200)



from __future__ import annotations

import datetime as dt
from typing import TYPE_CHECKING

import stripe
from django.db import models
from django.utils.translation import gettext_lazy as _

from ..compat import JSONField
from ..utils.request import get_client_ip
from .base import StripeModel

if TYPE_CHECKING:
    from django.http import HttpRequest


class WebhookEventQuerySet(models.QuerySet):
    def live(self) -> WebhookEventQuerySet:
        return self.filter(mode=WebhookEvent.Mode.LIVE)

    def test(self) -> WebhookEventQuerySet:
        return self.filter(mode=WebhookEvent.Mode.TEST)

    def received(self) -> WebhookEventQuerySet:
        return self.filter(state=WebhookEvent.State.RECEIVED)

    def processed(self) -> WebhookEventQuerySet:
        return self.filter(state=WebhookEvent.State.PROCESSED)


class WebhookEventManager(models.Manager):
    def create_from_event(
        self, request: HttpRequest, event: stripe.Event
    ) -> WebhookEvent:
        mode = WebhookEvent.Mode.from_stripe_livemode(event.livemode)
        stripe_created_at = dt.datetime.fromtimestamp(event.created)
        remote_ip = get_client_ip(request)

        stripe_request = event.get("request")
        request_id = stripe_request.id or ""
        request_idempotency_key = stripe_request.idempotency_key or ""

        return self.create(
            stripe_id=event.id,
            mode=mode,
            state=WebhookEvent.State.RECEIVED,
            event_type=event.type,
            data=event.data.object,
            api_version=event.api_version,
            stripe_created_at=stripe_created_at,
            request_id=request_id,
            request_idempotency_key=request_idempotency_key,
            headers=dict(request.headers),
            remote_ip=remote_ip,
        )


class WebhookEvent(StripeModel):
    """
    Models a single Stripe event received via the webhook.

    See https://stripe.com/docs/api/events/object

    Only webhook events with valid payloads & validated Stripe
    header signatures are stored.

    State: RECEIVED
    ---------------

    The webhook event has been received, and returned a success
    response to stripe - but no handlers were registered against
    the event type and thus no actual processing occurred.


    State: PROCESSED
    ----------------

    The webhook event has been received, and returned a success
    response to stripe, and was successfully processed by at least
    one registered handler.
    """

    class State(models.TextChoices):
        RECEIVED = ("RECEIVED", _("Received"))
        PROCESSED = ("PROCESSED", _("Processed"))

    state = models.CharField(
        max_length=20, choices=State.choices, default=State.RECEIVED
    )

    # Event-object related fields.
    event_type = models.CharField(max_length=100)
    data = JSONField(
        help_text=_(
            "The contents of data.object, representing the object "
            "the webhook event was triggered for."
        )
    )
    api_version = models.CharField(max_length=40)
    stripe_created_at = models.DateTimeField()

    # Instigating request related fields.
    request_id = models.CharField(max_length=255, blank=True)
    request_idempotency_key = models.CharField(max_length=255, blank=True)

    # Inbound webhook request data that we store
    # for security monitoring purposes only.
    headers = JSONField(blank=True)
    remote_ip = models.GenericIPAddressField(
        verbose_name=_("Remote IP"), help_text=_("IP address of the remote client.")
    )

    objects = WebhookEventManager.from_queryset(WebhookEventQuerySet)()

    def __str__(self) -> str:
        return f"WebhookEvent: {self.stripe_id}"

    def __repr__(self) -> str:
        return "<WebhookEvent id={} stripe_id={} mode={} state={}>".format(
            self.id, self.stripe_id, self.mode, self.state
        )

    @property
    def is_processed(self) -> bool:
        return self.state == self.State.PROCESSED

    def mark_processed(self) -> None:
        if self.is_processed:
            raise ValueError("You may not re-process a processed WebhookEvent")
        self.state = self.State.PROCESSED



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from django.http import HttpRequest


def get_client_ip(request: HttpRequest) -> Optional[str]:
    x_forwarded_for = request.headers.get("X-Forwarded-For")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[-1].strip()
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_stripe',
 'django_stripe.db',
 'django_stripe.migrations',
 'django_stripe.models',
 'django_stripe.utils']

package_data = \
{'': ['*']}

install_requires = \
['django>=3.0,<4.0', 'stripe>=2.48.0,<3.0.0']

setup_kwargs = {
    'name': 'django-stripe-lite',
    'version': '0.3',
    'description': 'A library to aid Django integration with Stripe.',
    'long_description': '# django-stripe-lite\n\nProvides a light-touch Django integration with Stripe.\n\nWe handle Stripe webhook security & persisting all events, while allowing your project to take care\nof the business logic.\n\nRequires PostgreSQL, Python 3.x & Django 3.x.\n\n## Installation & Usage\n\n```bash\npip install django-stripe-lite\n```\n\n**Include the app in your INSTALLED_APPS setting:**\n\n```python\nINSTALLED_APPS = (\n    ...,\n    "django_stripe",\n)\n```\n\n**Include the URLs in your URL conf:**\n\n```python\nfrom django.urls import include, path\n\nurlpatterns = [\n    # Assuming we\'re at the root, this will make the webhook\n    # available at /stripe/webhook/\n    path("stripe/", include("django_stripe.urls", namespace="stripe"))\n]\n```\n\n**Set the required settings in your settings file:**\n\n```python\nSTRIPE_WEBHOOK_SECRET = "whsec_0DoBceBjS0jjm7aQj459FXiFSluJEBxt"\n```\n\n**Run the migrations:**\n\n```bash\npython manage.py migrate django_stripe\n```\n\n**Set up your event handlers:**\n\nEvent handlers are simply functions in your code base, which are wrapped with a decorator which\nsignifies that they wish to handle a particular event type (or multiple) when it is received via the\nwebhook.\n\nAll event handlers must be imported at application startup, otherwise the decorator wil not be able\nto register them against the event type. An easy way to ensure this in your local project is to\ntrigger the import in one of your Django Apps `apps.py::AppConfig::ready()` method\n([see the docs](https://docs.djangoproject.com/en/3.0/ref/applications/#django.apps.AppConfig.ready)).\n\nWhen a webhook event is received, all processing of it is wrapped in a transaction such that a\nsingle event handler failure will result in an HTTP 500 returned from the endpoint and the\ntransaction will be rolled back resulting in no database changes for that request. This means that\nthe `WebhookEvent` is not persisted unless:\n\n-   it was received successfully and there were no active handlers registered for the event type,\n    or:\n-   it was received successfully and processed successfully by _all_ active handlers registered\n    against the event type.\n\n```python\nfrom django_stripe.models import WebhookEvent\nfrom django_stripe.webhooks import stripe_webhook_handler\n\n# Single event handler\n@stripe_webhook_handler("customer.subscription.deleted")\ndef delete_customer_subscription(event: WebhookEvent) -> Any:\n    # event.data (dict, Stripe Event object.data field, the object which triggered the webhook event)\n    # event.event_type (str, the full event type name e.g customer.subscription.deleted)\n    # event.mode (textchoices, LIVE or TEST)\n    # event.stripe_created_at (datetime, when Stripe created the event)\n    # event.db_created_at (datetime, when the database initially saved the event)\n    # event.db_last_updated_at (datetime, when the database last saved the event)\n    # event.stripe_id (str, Stripe Event ID)\n    # event.api_version (str, Stripe API version)\n    # event.request_id (str, the Stripe ID of the instigating request, if available)\n    # event.request_idempotency_key (str, the idempotency key of the instigating request, if available)\n    # event.is_processed (bool, whether the event was processed by a handler successfully)\n    # event.headers (dict, the headers of the webhook request)\n    # event.remote_ip (str, Remote IP of the webhook request)\n    pass\n\n# Multiple event handler\n@stripe_webhook_handler(\n    "customer.subscription.created",\n    "customer.subscription.deleted",\n    "customer.subscription.updated",\n)\ndef customer_handler(event: WebhookEvent) -> Any:\n    # See notes above for event structure.\n    pass\n```\n\nThat\'s it, you should be able to start receiving webhook events with the Stripe CLI test client.\nThen once you\'re ready, setup the production webhook via the Stripe dashboard.\n\n## Development\n\nCheck out the repo, then get the deps:\n\n```bash\npoetry install\n```\n\n## Tests\n\n#### Running tests\n\nThe tests themselves use `pytest` as the test runner. If you have installed the `poetry` evironment,\nyou can run them:\n\n```bash\n$ poetry run pytest\n```\n\nThe CI suite is controlled by `tox`, which contains a set of environments that will format (`fmt`),\nlint, and test against all supported Python + Django version combinations.\n\n```bash\n$ tox\n```\n\n#### CI\n\nCI is handled by GitHub Actions. See the Actions tab on Github & the `.github/workflows` folder.\n\n## Publish to PyPi\n\nUpdate versions, then:\n\n```bash\npoetry build\npoetry publish\n```\n',
    'author': 'YunoJuno',
    'author_email': 'code@yunojuno.com',
    'maintainer': 'YunoJuno',
    'maintainer_email': 'code@yunojuno.com',
    'url': 'https://github.com/yunojuno/django-stripe-lite',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
